#!/usr/bin/env python3

import os
import sys
import time
import random
import pprint
import _thread

import radical.utils as ru


try:
    # make pylint happy - this is optional code
    from Xlib import X, display, Xutil
except ImportError as exc:
    raise RuntimeError('need Xlib module to work') from exc



ROWS      =  1024
COLS      =  1024
CORES     =  ROWS * COLS  # number of cores to schedule over (sqrt must be int)
PPN       =    32         # cores per node, used for alignment
GPN       =     2         # GPUs per node, never aligned

ALIGN     =  True         # align small req onto single node
SCATTER   =  True         # allow scattered allocattions as fallback

CYCLES    = 10000         # number of cycles
CPU_MIN   =     0         # minimal number of cores requested
CPU_MAX   =    32         # maximal number of cores requested
GPU_MIN   =     0         # minimal number of GPUs  requested
GPU_MAX   =     4         # maximal number of GPUs  requested
REQ_BULK  =     2         # number of requests to handle in bulk
REL_PROB  =     0.010     # probablility of release per cycle
REQ_MIN   =     1         # probablility of release per cycle
REQ_MAX   =     2         # probablility of release per cycle


# ------------------------------------------------------------------------------
#
class SchedulerViz(object):

    # --------------------------------------------------------------------------
    #
    def __init__(self, args):

        self.scheduler = args[0]
        self.cycles    = args[1]
        self.req_min   = args[2]
        self.req_max   = args[3]
        self.req_bulk  = args[4]
        self.rel_prob  = args[5]

        print(scheduler)

        self.d      = display.Display()
        self.screen = self.d.screen()
        self.window = self.screen.root.create_window(
            0, 0, ROWS, COLS, 2,
            self.screen.root_depth,
            X.InputOutput,
            X.CopyFromParent,

            background_pixel=self.screen.white_pixel,

            event_mask=(X.ExposureMask        |
                        X.StructureNotifyMask |
                        X.ButtonPressMask     |
                        X.ButtonReleaseMask   |
                        X.Button1MotionMask   ),
            colormap=X.CopyFromParent)

        self.gc = self.window.create_gc(foreground=self.screen.black_pixel,
                                        background=self.screen.white_pixel)

        self.WM_DELETE_WINDOW = self.d.intern_atom('WM_DELETE_WINDOW')
        self.WM_PROTOCOLS     = self.d.intern_atom('WM_PROTOCOLS')

        self.window.set_wm_name     ('ru_scheduler viz')
        self.window.set_wm_icon_name('ru_scheduler')
        self.window.set_wm_class    ('ru_scheduler', 'XlibExample')

        self.window.set_wm_protocols([self.WM_DELETE_WINDOW])
        self.window.set_wm_hints(flags=Xutil.StateHint,
                                 initial_state=Xutil.NormalState)

        self.window.set_wm_normal_hints(flags=(Xutil.PPosition | Xutil.PSize |
                                               Xutil.PMinSize),
                                        min_width=20,
                                        min_height=20)
        self.window.map()

        self._thread = _thread.start_new_thread(self.update_viz,   tuple())
        self._thread = _thread.start_new_thread(self.update_sched, tuple())


    # --------------------------------------------------------------------------
    # Main loop, handling events
    def loop(self):

        while 1:
            e = self.d.next_event()

            if e.type == X.DestroyNotify:
              # self.term.set()
                sys.exit(0)

            if e.type == X.ButtonPress and e.detail == 1:
              # self.term.set()
                sys.exit(0)

            if e.type == X.ClientMessage:
                if e.client_type == self.WM_PROTOCOLS:
                    fmt, data = e.data
                    if fmt == 32 and data[0] == self.WM_DELETE_WINDOW:
                      # self.term.set()
                        sys.exit(0)


    # --------------------------------------------------------------------------
    #
    def update_viz(self):

        # max len for point list us USHORT_MAX, ie. 64k
        chunk = 128 * 128
        cmap  = self.d.screen().default_colormap
        red   = cmap.alloc_named_color("red" ).pixel
        blue  = cmap.alloc_named_color("blue").pixel

        while True:

            print(1)

            state   = self.scheduler.get_map()
            active  = list()
            passive = list()

            active_chunk  = list()
            passive_chunk = list()

            for r in range(ROWS):
                for c in range(COLS):
                    if state[r * COLS + c]:
                        active_chunk.append((r, c))
                        if len(active_chunk) > chunk:
                            active.append(active_chunk)
                            active_chunk = list()
                    else:
                        passive_chunk.append((r, c))
                        if len(passive_chunk) > chunk:
                            passive.append(passive_chunk)
                            passive_chunk = list()

            if active_chunk:
                active.append(active_chunk)
            if passive_chunk:
                passive.append(passive_chunk)

            print(2)

            self.gc.change(foreground=red)
            for active_chunk in active:
                self.window.poly_point(self.gc, X.CoordModeOrigin, active_chunk)

            self.gc.change(foreground=blue)
            for passive_chunk in passive:
                self.window.poly_point(self.gc, X.CoordModeOrigin,passive_chunk)
            print(3)

            self.d.flush()
            print('done')
            time.sleep(0.1)


    # --------------------------------------------------------------------------
    #
    def update_sched(self):

        try:
            # ------------------------------------------------------------------
            #
            # This implementation will first create a number of requests,
            # specifically 'REQ_BULK' requests, each randomly distributed
            # between 'REQ_MIN' and 'REQ_MAX'.  Those requests are the
            # scheduled via calls to 'self.scheduler.allocate(req)', and the
            # results are stored in a 'running' list.
            #
            # After that requested bulk is allocated, all items in 'running'
            # (including those allocated in earlier self.cycles)  are up for
            # release, with a certain probaility REL_PROB.  For example, for
            # 'REL_PROB = 0.01',  approximately 1% of the entries will be
            # released, via a call to 'self.scheduler.deallocate(res)'.
            #
            # The above cycle repeats 'self.cycles' times, with a (potentially)
            # ever increasing 'running' list, The load on the self.scheduler is
            # thus continuously increasing as cores remain allocated over
            # self.cycles, and the self.scheduler needs to search harder for
            # free cores
            #
            # This thread finishes when:
            #   - all self.cycles are completed
            #   - the self.scheduler cannot find anough cores for a request
            #   - the self.scheduler can't align an allocation (if 'ALIGN' is
            #     set)
            #
            # ------------------------------------------------------------------

            running = list()
            done    = list()

            total_start   = time.time()
            total_alloc   = 0
            total_dealloc = 0
            total_align   = 0
            total_scatter = 0


            # for range 1024:
            #   find 1024 chunks of 16  cores
            #   free  512 chunks of  8 or 16 cores (random)
            abort_cycles = False
            for cycle in range(self.cycles):

                time.sleep(0.1)

                if abort_cycles:
                    break

                # we randomly request cores in a certain range
                requests = list()
                for _ in range(self.req_bulk):
                    requests.append(random.randint(self.req_min,self.req_max))

                tmp = list()
                start = time.time()
                try:
                    for req in requests:
                        tmp.append(self.scheduler.alloc(req))
                except Exception as e:
                    print(e)
                    abort_cycles = True
                stop = time.time()

                for res in tmp:
                    total_alloc += 1
                    if res[2]:
                        total_scatter += 1
                    if res[3]:
                        total_align += 1
                running += tmp

                if (stop == start):
                    alloc_rate = -1
                else:
                    alloc_rate = len(requests) / (stop - start)

                if abort_cycles:
                    # don't dealloc, as it screws with statistics
                    dealloc_rate = -1

                else:

                    # build a list of release candidates and, well, release them
                    to_release = list()
                    for idx in reversed(list(range(len(running)))):
                        if random.random() < self.rel_prob:
                            to_release.append(running[idx])
                            del(running[idx])

                    start = time.time()
                    try:
                        for res in to_release:
                            self.scheduler.dealloc(res)
                            done.append(res)
                    except Exception as e:
                        print(e)
                        abort_cycles = True
                    stop   = time.time()
                    total_dealloc += len(to_release)

                    if (stop == start):
                        dealloc_rate = -1
                    else:
                        dealloc_rate = len(to_release) / (stop - start)

                print('%5d : alloc : %6d (%8.1f/s)   dealloc : %6d (%8.1f/s)'
                      'free %6d' % (cycle, total_alloc,   alloc_rate,
                                           total_dealloc, dealloc_rate,
                                           self.scheduler.get_map.count()))

            if abort_cycles:
                print('cycle aborted')
            else:
                print('cycles done')

        except Exception:
            import traceback
            print(traceback.format_exc(sys.exc_info()))

        total_stop = time.time()
        stats = self.scheduler.get_stats()

        # NOTE: Uncomment to DEBUG
        # continuous stretches of #free/busy cores
        # print()
        # print('\ncores :  free :  busy')
        # counts = set(list(stats['free_dist'].keys()) + \
        #          list(stats['busy_dist'].keys()))
        # for count in sorted(counts):
        #     print('%5d : %5s : %5s' % (count,
        #             stats['free_dist'].get(count, ''),
        #             stats['busy_dist'].get(count, '')))

        # distributions of free cores over nodes
        print()
        print('free : nodes')
        for i in sorted(stats['node_free'].keys()):
            print(' %3d : %5d' % (i, stats['node_free'].get(i, '')))

        print()
        print('total cores  : %9d' % stats['total'])
        print('      free   : %9d' % stats['free'])
        print('      busy   : %9d' % stats['busy'])
        print('      alloc  : %9d' % total_alloc)
        print('      align  : %9d' % total_align)
        print('      scatter: %9d' % total_scatter)
        print('      dealloc: %9d' % total_dealloc)
        print('      runtime: %9.1fs'  % (total_stop   - total_start))
        print('      ops/sec: %9.1f/s' % ((total_alloc + total_dealloc) /
                                          (total_stop  - total_start)))

        # NOTE: Uncomment to DEBUG
        # idx = 0
        # with open('cores', 'w') as f:
        #     node = '%5d : ' % idx
        #     for b in self.scheduler.get_map():
        #         if not idx % ppn:
        #             node += '\n'
        #         if b:
        #             node += '#'
        #         else:
        #             node += ' '
        #         idx += 1
        #     f.write(node)

        print('\nuse <Esc> in viz-window to quit\n')


# ------------------------------------------------------------------------------
#
if __name__ == "__main__":

    if len(sys.argv) >= 3:

        if '/' not in __file__:
            path = '.'
        elif __file__:
            path = os.path.dirname(__file__)
        else:
            path = os.path.dirname(sys.argv[0])
        config = ru.read_json("%s/radical-utils-scheduler.json" % path)

        cluster_id  = sys.argv[1]
        workload_id = sys.argv[2]

        cluster  = config['cluster'][cluster_id]
        workload = config['workload'][workload_id]

        print('using cluster: %s'  % pprint.pformat(cluster))
        print('using workload: %s' % pprint.pformat(workload))

        cores   =  int(cluster['cores'])
        ppn     =  int(cluster['ppn'])
        align   = bool(cluster['align'])
        scatter = bool(cluster['scatter'])

        cycles   =   int(workload['cycles'])
        req_min  =   int(workload['req_min'])
        req_max  =   int(workload['req_max'])
        req_bulk =   int(workload['req_bulk'])
        rel_prob = float(workload['rel_prob'])

    else:
        cores   = CORES
        ppn     = PPN
        align   = ALIGN
        scatter = SCATTER

        cycles   = CYCLES
        req_min  = REQ_MIN
        req_max  = REQ_MAX
        req_bulk = REQ_BULK
        rel_prob = REL_PROB


    scheduler = ru.scheduler.BitarrayScheduler({'cores'   : cores,
                                                'ppn'     : ppn,
                                                'align'   : align,
                                                'scatter' : scatter})

    vs_args = [scheduler, cycles, req_min, req_max, req_bulk, rel_prob]
    vs      = SchedulerViz(vs_args)
    vs.loop()


# ------------------------------------------------------------------------------





__author__    = 'RADICAL-Cybertools Team'
__copyright__ = 'Copyright 2021, The RADICAL-Cybertools Team'
__license__   = 'MIT'

import collections
import math
import netifaces
import socket

from functools import reduce

from .misc   import as_list, ru_open
from .config import DefaultConfig


# ------------------------------------------------------------------------------
#
_hostname = None


def get_hostname():
    """Look up the hostname."""

    global _hostname                                     # pylint: disable=W0603
    if not _hostname:

        _hostname = socket.getfqdn()

    return _hostname


# ------------------------------------------------------------------------------
#
def get_hostip(req=None, log=None):
    """Look up the IP address for a given requested interface name.
    If interface is not given, do some magic."""

    AF_INET = netifaces.AF_INET

    # We create an ordered preference list, consisting of:
    #   - given arglist
    #   - white list (hardcoded preferred interfaces)
    #   - black_list (hardcoded unfavorable interfaces)
    #   - all others (whatever is not in the above)
    # Then this list is traversed, we check if the interface exists and has an
    # IP address.  The first match is used.

    if not req:
        req = DefaultConfig().get('iface')

    req = as_list(req)

    white_list = [
        'ens1f1',   # amarel
        'ib0',      # infiniband   # NOTE: unusable on Amarel, use `req` there!
        'hsn0',     # Frontier (HPE Cray EX)
        'ipogif0',  # Cray's
        'br0',      # SuperMIC
        'eth0',     # desktops etc.
        'wlan0'     # laptops etc.
    ]

    black_list = [
        'lo',       # takes the 'inter' out of the 'net'
        'sit0'      # ?
    ]

    ifaces = netifaces.interfaces()
    rest   = [iface for iface in ifaces
                     if iface not in req        and
                        iface not in white_list and
                        iface not in black_list]

    preflist = req + white_list + rest

    for iface in preflist:

        if iface not in ifaces:
            if log:
                log.debug('check iface %s: does not exist', iface)
            continue

        info = netifaces.ifaddresses(iface)
        if AF_INET not in info:
            if log:
                log.debug('check iface %s: no information', iface)
            continue

        if not len(info[AF_INET]):
            if log:
                log.debug('check iface %s: insufficient information', iface)
            continue

        if not info[AF_INET][0].get('addr'):
            if log:
                log.debug('check iface %s: disconnected', iface)
            continue

        ip = info[AF_INET][0].get('addr')
        if log:
            log.debug('check iface %s: ip is %s', iface, ip)

        if ip:
            return ip

    return '127.0.0.1'


# ------------------------------------------------------------------------------
#
def create_hostfile(sandbox, name, hostlist, sep=' ', impaired=False):

    hostlist = as_list(hostlist)
    filename = '%s/%s.hosts' % (sandbox or '.', name)
    with ru_open(filename, 'w', encoding='utf8') as fout:

        if not impaired:
            # create dict: {'host1': x, 'host2': y}
            counter = collections.Counter(hostlist)
            # convert it into an ordered dict,
            count_dict = collections.OrderedDict(sorted(counter.items(),
                                                        key=lambda h: h[0]))

            hosts = ['%s%s%d' % (h, sep, c) for h, c in count_dict.items()]
            fout.write('\n'.join(hosts) + '\n')

        else:
            # write entry "hostN\nhostM\n"
            fout.write('\n'.join(hostlist) + '\n')

    # return the filename, caller is responsible for cleaning up
    return filename


# ------------------------------------------------------------------------------
#
def compress_hostlist(hostlist):

    # create dict: {'host1': x, 'host2': y}
    count_dict = dict(collections.Counter(hostlist))
    # find the gcd of the host counts (gcd of a list of numbers)
    host_gcd = reduce(math.gcd, set(count_dict.values()))

    # divide host counts by the gcd
    for host in count_dict:
        count_dict[host] /= host_gcd

    # recreate a list of hosts based on the normalized dict
    hosts = []
    for host, count in count_dict.items():
        hosts.extend([host] * int(count))

    # sort the list for readability
    hosts.sort()

    return hosts


# ------------------------------------------------------------------------------
#
def get_hostlist_by_range(hoststring, prefix='', width=0):
    """Convert string with host IDs into list of hosts.

    Example: Cobalt RM would have host template as 'nid%05d'
                get_hostlist_by_range('1-3,5', prefix='nid', width=5) =>
                ['nid00001', 'nid00002', 'nid00003', 'nid00005']
    """

    if not hoststring.replace('-', '').replace(',', '').isnumeric():
        raise ValueError('non numeric set of ranges (%s)' % hoststring)

    host_ids = []
    id_width = 0

    for num in hoststring.split(','):
        num_range = num.split('-')

        if len(num_range) > 1:
            num_lo, num_hi = num_range
            if not num_lo or not num_hi:
                raise ValueError('incorrect range format (%s)' % num)
            host_ids.extend(list(range(int(num_lo), int(num_hi) + 1)))

        else:
            host_ids.append(int(num_range[0]))

        id_width = max(id_width, *[len(n) for n in num_range])

    width = width or id_width
    return ['%s%0*d' % (prefix, width, hid) for hid in host_ids]


# ------------------------------------------------------------------------------
#
def get_hostlist(hoststring):
    """Convert string with hosts (IDs within brackets) into list of hosts.

    Example: 'node-b1-[1-3,5],node-c1-4,node-d3-3,node-k[10-12,15]' =>
             ['node-b1-1', 'node-b1-2', 'node-b1-3', 'node-b1-5',
              'node-c1-4', 'node-d3-3',
              'node-k10', 'node-k11', 'node-k12', 'node-k15']
    """

    output = []

    hoststring += ','
    host_group  = []

    idx, idx_stop = 0, len(hoststring)
    while idx != idx_stop:

        comma_idx   = hoststring.find(',', idx)
        bracket_idx = hoststring.find('[', idx)

        if comma_idx >= 0 and (bracket_idx == -1 or comma_idx < bracket_idx):

            if host_group:
                prefix = hoststring[idx:comma_idx]
                if prefix:
                    for h_idx in range(len(host_group)):
                        host_group[h_idx] += prefix
                output.extend(host_group)
                del host_group[:]

            else:
                output.append(hoststring[idx:comma_idx])

            idx = comma_idx + 1

        elif bracket_idx >= 0 and (comma_idx == -1 or bracket_idx < comma_idx):

            prefix = hoststring[idx:bracket_idx]
            if not host_group:
                host_group.append(prefix)
            else:
                for h_idx in range(len(host_group)):
                    host_group[h_idx] += prefix

            closed_bracket_idx = hoststring.find(']', bracket_idx)
            range_set = hoststring[(bracket_idx + 1):closed_bracket_idx]

            host_group_ = []
            for prefix in host_group:
                host_group_.extend(get_hostlist_by_range(range_set, prefix))
            host_group = host_group_

            idx = closed_bracket_idx + 1

    return output


# --------------------------------------------------------------------------
#
def is_localhost(host: str) -> bool:
    '''
    Returns `True` if given hostname is localhost, `False` otherwise.
    '''

    if not host:
        return True

    elif host == 'localhost':
        return True

    else:
        sockhost = socket.gethostname()
        while sockhost:
            if host == sockhost:
                return True
            sockhost = '.'.join(sockhost.split('.')[1:])

    return False


# ------------------------------------------------------------------------------





import zmq

import threading as mt

from typing      import Optional, Union, Iterator, Any, Dict

from ..ids       import generate_id
from ..url       import Url
from ..misc      import as_string
from ..host      import get_hostip
from ..logger    import Logger
from ..profile   import Profiler
from ..debug     import get_exception_trace
from ..serialize import to_msgpack, from_msgpack

from .utils      import no_intr


# --------------------------------------------------------------------------
#
_LINGER_TIMEOUT    =         250  # ms to linger after close
_HIGH_WATER_MARK   = 1024 * 1024  # number of messages to buffer before dropping
_DEFAULT_BULK_SIZE =        1024  # number of messages to put in a bulk


# ------------------------------------------------------------------------------
#
class Server(object):

    # --------------------------------------------------------------------------
    #
    def __init__(self, url:  Optional[str] = None,
                       uid:  Optional[str] = None,
                       path: Optional[str] = None) -> None:

        # this server offers only synchronous communication: a request will be
        # worked upon and answered before the next request is received.

        self._url  = url
        self._cbs  = dict()
        self._path = path

        if not self._path:
            self._path = './'

        if uid: self._uid = uid
        else  : self._uid = generate_id('server', ns='radical.utils')

        self._log    = Logger(self._uid,   path=self._path)
        self._prof   = Profiler(self._uid, path=self._path)

        self._addr   = None
        self._thread = None
        self._up     = mt.Event()
        self._term   = mt.Event()

        self.register_request('echo', self._request_echo)
        self.register_request('fail', self._request_fail)

        if not self._url:
            self._url = 'tcp://*:10000-11000'

        # URLs can specify port ranges to use - check if that is the case (see
        # default above) and initilize iterator.  The URL is expected to have
        # the form:
        #
        #   <proto>://<iface>:<ports>/
        #
        # where
        #   <proto>: any protocol accepted by zmq,       defaults to `tcp`
        #   <iface>: IP number of interface to bind to   defaults to `*`
        #   <ports>: port range to find port to bind to  defaults to `*`
        #
        # The port range can be formed as:
        #
        #   '*'      : any port
        #   '100+'   : any port equal or larger than 100
        #   '100-'   : any port equal or larger than 100
        #   '100-110': any port equal or larger than 100, up to 110
        tmp = self._url.split(':', 2)
        assert len(tmp) == 3
        self._proto = tmp[0]
        self._iface = tmp[1].lstrip('/')
        self._ports = tmp[2].replace('+', '-')

        tmp = self._ports.split('-')

        self._port_this : Union[int, str, None] = None
        self._port_start: Optional[int]
        self._port_stop : Optional[int]

        if len(tmp) == 0:
            self._port_start = 1
            self._port_stop  = None
        elif len(tmp) == 1:
            if tmp[0] == '*':
                self._port_this  = '*'
                self._port_start = None
                self._port_stop  = None
            else:
                self._port_start = int(tmp[0])
                self._port_stop  = int(tmp[0])
        elif len(tmp) == 2:
            if tmp[0]: self._port_start = int(tmp[0])
            else     : self._port_start = 1
            if tmp[1]: self._port_stop  = int(tmp[1])
            else     : self._port_stop  = None
        else:
            raise RuntimeError('cannot parse port spec %s' % self._ports)


    # --------------------------------------------------------------------------
    #
    def _iterate_ports(self) -> Iterator[Union[int, str, None]]:

        if self._port_this == '*':
            # leave scanning to zmq
            yield self._port_this

        if self._port_this is None:
            # initialize range iterator
            self._port_this = self._port_start

        if self._port_stop is None:
            while True:
                yield self._port_this
                self._port_this += 1

        else:
            # make type checker happy
            assert isinstance(self._port_this,  int)
            assert isinstance(self._port_start, int)

            while self._port_this <= self._port_stop:
                yield self._port_this
                self._port_this += 1


    # --------------------------------------------------------------------------
    #
    def _iterate_urls(self) -> Iterator[str]:

        for port in self._iterate_ports():
            yield '%s://%s:%s' % (self._proto, self._iface, port)


    # --------------------------------------------------------------------------
    #
    @property
    def uid(self) -> str:

        return self._uid


    @property
    def addr(self) -> Optional[str]:

        return self._addr


    # --------------------------------------------------------------------------
    #
    def start(self) -> None:

        self._log.info('start bridge %s', self._uid)

        if self._thread:
            raise RuntimeError('`start()` can be called only once')

        self._thread = mt.Thread(target=self._work)
        self._thread.daemon = True
        self._thread.start()

        self._up.wait()


    # --------------------------------------------------------------------------
    #
    def stop(self) -> None:

        self._log.info('stop bridge %s', self._uid)
        self._term.set()


    # --------------------------------------------------------------------------
    #
    def wait(self) -> None:

        self._log.info('wait bridge %s', self._uid)

        if self._thread:
            self._thread.join()

        self._log.info('wait bridge %s', self._uid)


    # --------------------------------------------------------------------------
    #
    def register_request(self, req, cb) -> None:

        self._log.info('add handler: %s: %s', req, cb)
        self._cbs[req] = cb


    # --------------------------------------------------------------------------
    #
    def _request_fail(self, arg) -> None:

        raise RuntimeError('task failed successfully')


    # --------------------------------------------------------------------------
    #
    def _request_echo(self, arg: Any) -> Any:

        return arg


    # --------------------------------------------------------------------------
    #
    def _success(self, res: Optional[str] = None) -> Dict[str, Optional[str]]:

        return {'err': None,
                'exc': None,
                'res': res}


    # --------------------------------------------------------------------------
    #
    def _error(self, err: Optional[str] = None,
                     exc: Optional[str] = None) -> Dict[str, Optional[str]]:

        if not err:
            err = 'invalid request'

        return {'err': err,
                'exc': exc,
                'res': None}


    # --------------------------------------------------------------------------
    #
    def _work(self) -> None:

        self._ctx  = zmq.Context()
        self._sock = self._ctx.socket(zmq.REP)

        self._sock.linger = _LINGER_TIMEOUT
        self._sock.hwm    = _HIGH_WATER_MARK

        for url in self._iterate_urls():
            try:
                self._log.debug('try url %s', url)
                self._sock.bind(url)
                self._log.debug('success')
                break
            except zmq.error.ZMQError as e:
                if 'Address already in use' in str(e):
                    self._log.warn('port in use - try next (%s)' % url)
                else:
                    raise

        addr       = Url(as_string(self._sock.getsockopt(zmq.LAST_ENDPOINT)))
        addr.host  = get_hostip()
        self._addr = str(addr)

        self._up.set()

        self._poll = zmq.Poller()
        self._poll.register(self._sock, zmq.POLLIN)

        while not self._term.is_set():

            event = dict(no_intr(self._poll.poll, timeout=100))

            if self._sock not in event:
                continue

            # default response
            rep = None
            req = None

            try:
                data = no_intr(self._sock.recv)
                req  = as_string(from_msgpack(data))
                self._log.debug('req: %s', str(req)[:128])

                if not isinstance(req, dict):
                    rep = self._error(err='invalid message type')

                else:
                    cmd    = req['cmd']
                    args   = req['args']
                    kwargs = req['kwargs']

                    if not cmd:
                        rep = self._error(err='no command in request')

                    elif cmd not in self._cbs:
                        rep = self._error(err='command [%s] unknown' % cmd)

                    else:
                        rep = self._success(self._cbs[cmd](*args, **kwargs))

            except Exception as e:
                self._log.exception('command failed: %s', req)
                rep = self._error(err='command failed: %s' % str(e),
                                  exc='\n'.join(get_exception_trace()))

            finally:
                if not rep:
                    rep = self._error('server error')
                no_intr(self._sock.send, to_msgpack(rep))
                self._log.debug('rep: %s', str(rep)[:128])

        self._sock.close()
        self._log.debug('term')


# ------------------------------------------------------------------------------





import json
import msgpack

from .typeddict import as_dict, TypedDict

# ------------------------------------------------------------------------------
#
class _ClassType:

    def __init__(self, ctype, encode, decode):

        self.ctype : type     = ctype
        self.encode: callable = encode
        self.decode: callable = decode


_ctypes = dict()


# ------------------------------------------------------------------------------
#
def register_serializable(cls, encode=None, decode=None):
    '''
    register a class for json and msgpack serialization / deserialization.

    Args:
        cls    (type):     class type to register
        encode (callable): converts class instance into encodable data structure
        decode (callable): recreates the class instance from that data structure
    '''

    if encode is None: encode = cls
    if decode is None: decode = cls

    _ctypes[cls.__name__] = _ClassType(cls, encode, decode)

register_serializable(TypedDict)


# ------------------------------------------------------------------------------
#
class _json_encoder(json.JSONEncoder):
    '''
    internal methods to encode registered classes to json
    '''

    def encode(self, o, *args, **kw):
        tmp = as_dict(o, _annotate=True)
        return super().encode(tmp, *args, **kw)

    def default(self, o):
        for cname,methods in _ctypes.items():
            if isinstance(o, methods.ctype):
                return {'_type': cname,
                        'as_str': methods.encode(o)}
        return super().default(o)


# ------------------------------------------------------------------------------
#
def _json_decoder(obj):
    '''
    internal methods to decode registered classes from json
    '''
    otype = obj.get('_type')
    if not otype:
        return obj

    methods = _ctypes.get(otype)
    if not methods:
        return obj

    del obj['_type']
    if 'as_str' in obj:
        return methods.decode(obj['as_str'])

    return methods.decode(obj)


# ------------------------------------------------------------------------------
#
def _msgpack_encoder(obj):
    '''
    internal methods to encode registered classes to msgpack
    '''
    for cname,methods in _ctypes.items():
        if isinstance(obj, methods.ctype):
            return {'__%s__' % cname: True, 'as_str': methods.encode(obj)}
    return obj


# ------------------------------------------------------------------------------
#
def _msgpack_decoder(obj):
    '''
    internal methods to decode registered classes from msgpack
    '''
    for cname,methods in _ctypes.items():
        if '__%s__' % cname in obj:
            return methods.decode(obj['as_str'])
    return obj


# ------------------------------------------------------------------------------
#
def to_json(data):
    '''
    convert data to json, using registered classes for serialization

    Args:
        data (object): data to be serialized

    Returns:
        str: json serialized data
    '''
    return json.dumps(data, sort_keys=True, indent=4, ensure_ascii=False,
                            cls=_json_encoder)


# ------------------------------------------------------------------------------
#
def from_json(data):
    '''
    convert json data to python data structures, using registered classes for
    deserialization

    Args:
        data (str): json data to be deserialized

    Returns:
        object: deserialized data
    '''
    return json.loads(data, object_hook=_json_decoder)


# ------------------------------------------------------------------------------
#
def to_msgpack(data):
    '''
    convert data to msgpack, using registered classes for serialization

    Args:
        data (object): data to be serialized

    Returns:
        bytes: msgpack serialized data
    '''
    return msgpack.packb(data, default=_msgpack_encoder, use_bin_type=True)


# ------------------------------------------------------------------------------
#
def from_msgpack(data):
    '''
    convert msgpack data to python data structures, using registered classes for
    deserialization

    Args:
        data (bytes): msgpack data to be deserialized

    Returns:
        object: deserialized data
    '''
    return msgpack.unpackb(data, object_hook=_msgpack_decoder,
                           raw=False, strict_map_key=False)


# ------------------------------------------------------------------------------




#!/usr/bin/env python 

import zmq
import time

addr = None
with open('test.bridge.url', 'r') as fin:
    for line in fin.readlines():
        tag, addr = line.split()
        if tag == 'PUB':
            break

print('add: %s' % addr)

context = zmq.Context()
socket  = context.socket(zmq.PUB)
socket.connect(addr)

n     = 300
start = time.time()
topic = 'topic'
for index in range(n):
    socket.send_multipart([topic, '-'])
    socket.send_multipart([topic, '+'])

stop = time.time()

socket.send_multipart([topic, 'x'])
print('-> %.2f /s' % ((n * 2) / (stop - start)))




# Copyright 2009 Google Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# Licensed to the PSF under a Contributor Agreement.
#
# Author: Gregory P. Smith <greg@krypto.org>

'''
This module implements a pthread_atfork() work-a-like mechanism for all
fork() calls made from the Python os module.  Any time a fork() is called
from Python a set of unique callbacks will be made in each of the following
three states:

  - Preparing to fork        - Immediately before the fork call is made.
  - In the child after fork  - Immediately after the fork in the child process.
  - In the parent after fork - Immediately after the fork in the parent process.

NOTE: The `parent` callback will executed also if the fork failed.

To use this module, first import it early on your programs initialization::

    import atfork
    atfork.monkeypatch_os_fork_functions()

That will stub out os.fork and os.forkpty with wrapped versions implementing
the enhanced behavior.

Next, register your atfork actions by calling atfork.atfork::

    atfork.atfork(prepare=my_lock.acquire,
                  parent=my_lock.release,
                  child=my_lock.release)

No API to unregister an atfork call is provided.  If you are concerned
about resource usage by references your callable holds, consider using
weakref's within your callable.
'''

import os
import sys
import threading
import traceback

_orig_os_fork    = os.fork
_orig_os_forkpty = os.forkpty


# ------------------------------------------------------------------------------
#
def monkeypatch_os_fork_functions():
    '''
    Replace os.fork* with wrappers that use ForkSafeLock to acquire
    all locks before forking and release them afterwards.
    '''

    # monkeypatching can be enabled by setting RADICAL_UTILS_NO_ATFORK
    if 'RADICAL_UTILS_NO_ATFORK' in os.environ:
        return

    builtin_function = type(''.join)
    if hasattr(os, 'fork') and isinstance(os.fork, builtin_function):
        os.fork = os_fork_wrapper
    if hasattr(os, 'forkpty') and isinstance(os.forkpty, builtin_function):
        os.forkpty = os_forkpty_wrapper


# This lock protects all of the lists below.
_fork_lock               = threading.RLock()
_prepare_call_list       = list()
_prepare_call_exceptions = list()
_parent_call_list        = list()
_child_call_list         = list()


# ------------------------------------------------------------------------------
#
def atfork(prepare=None, parent=None, child=None):
    '''
    A Python work-a-like of pthread_atfork.

    Any time a fork() is called from Python, all 'prepare' callables will
    be called in the order they were registered using this function.

    After the fork (successful or not), all 'parent' callables will be called in
    the parent process.  If the fork succeeded, all 'child' callables will be
    called in the child process.

    No exceptions may be raised from any of the registered callables.  If so
    they will be printed to sys.stderr after the fork call once it is safe
    to do so.
    '''

    # monkeypatching can be disabled by setting RADICAL_UTILS_NO_ATFORK
    if 'RADICAL_UTILS_NO_ATFORK' in os.environ:
        return

    assert not prepare or callable(prepare)
    assert not parent or callable(parent)
    assert not child or callable(child)

    _fork_lock.acquire()

    try:
        if prepare:
            _prepare_call_list.append(prepare)
        if parent:
            _parent_call_list.append(parent)
        if child:
            _child_call_list.append(child)
    finally:
        try:
            _fork_lock.release()
        except:
            pass


# ------------------------------------------------------------------------------
#
def _call_atfork_list(call_list):
    '''
    Given a list of callables in call_list, call them all in order and save
    and return a list of sys.exc_info() tuples for each exception raised.
    '''
    exception_list = []
    for func in call_list:
        try:
            try:
                func()
            except:
                pass
        except:
            exception_list.append(sys.exc_info())
    return exception_list


# ------------------------------------------------------------------------------
#
def prepare_to_fork_acquire():
    '''Acquire our lock and call all prepare callables.'''
    _fork_lock.acquire()
    _prepare_call_exceptions.extend(_call_atfork_list(_prepare_call_list))


# ------------------------------------------------------------------------------
#
def parent_after_fork_release():
    '''
    Call all parent after fork callables, release the lock and print
    all prepare and parent callback exceptions.
    '''

    prepare_exceptions = list(_prepare_call_exceptions)
    del _prepare_call_exceptions[:]
    exceptions = _call_atfork_list(_parent_call_list)
    try:
        _fork_lock.release()
    except:
        pass

    _print_exception_list(prepare_exceptions, 'before fork')
    _print_exception_list(exceptions, 'after fork from parent')


# ------------------------------------------------------------------------------
#
def child_after_fork_release():
    '''
    Call all child after fork callables, release lock and print all
    all child callback exceptions.
    '''

    del _prepare_call_exceptions[:]
    exceptions = _call_atfork_list(_child_call_list)
    try:
        _fork_lock.release()
    except:
        pass

    _print_exception_list(exceptions, 'after fork from child')


# ------------------------------------------------------------------------------
#
def _print_exception_list(exceptions, message, output_file=None):
    '''
    Given a list of sys.exc_info tuples, print them all using the traceback
    module preceeded by a message and separated by a blank line.
    '''

    output_file = output_file or sys.stderr
    message = 'Exception %s:\n' % message
    for exc_type, exc_value, exc_traceback in exceptions:
        output_file.write(message)
        traceback.print_exception(exc_type, exc_value, exc_traceback,
                                  file=output_file)
        output_file.write('\n')


# ------------------------------------------------------------------------------
#
def os_fork_wrapper():
    '''Wraps os.fork() to run atfork handlers.'''
    pid = None
    prepare_to_fork_acquire()
    try:
        pid = _orig_os_fork()
    finally:
        if pid == 0:
            child_after_fork_release()
        else:
            # We call this regardless of fork success in order for
            # the program to be in a sane state afterwards.
            parent_after_fork_release()
    return pid


# ------------------------------------------------------------------------------
#
def os_forkpty_wrapper():
    '''Wraps os.forkpty() to run atfork handlers.'''
    pid = None
    prepare_to_fork_acquire()
    try:
        pid, fd = _orig_os_forkpty()
    finally:
        if pid == 0:
            child_after_fork_release()
        else:
            parent_after_fork_release()
    return pid, fd


# TODO: Also replace os.fork1() on Solaris.

# ------------------------------------------------------------------------------





__author__    = 'Radical.Utils Development Team (Andre Merzky, Ole Weidner)'
__copyright__ = 'Copyright 2013, RADICAL@Rutgers'
__license__   = 'MIT'


import os

from .contrib import urlparse25 as urlparse
from .        import signatures as rus


# ------------------------------------------------------------------------------
#
class Url(object):
    '''
    The RADICAL Url class.

    URLs are used in several places in the RADICAL software projects: to specify
    service endpoints for job submission or resource management, for file or
    directory locations, etc.

    The URL class is designed to simplify URL management for these purposes --
    it allows to manipulate individual URL elements, while ensuring that the
    resulting URL is well formatted. Example::

      # create a URL from a string
      location = radical.utils.Url('file://localhost/tmp/file.dat')
      d = radical.utils.filesystem.Directory(location)

    A URL consists of the following components (where one ore more can be
    'None')::

      <scheme>://<user>:<pass>@<host>:<port>/<path>?<query>#<fragment>

    Each of these components can be accessed via its property or alternatively,
    via getter / setter methods. Example::

      url = ru.Url('scheme://pass:user@host:port/path?query#fragment')

      # modify the scheme
      url.scheme = 'anotherscheme'

      # above is equivalent with
      url.set_scheme('anotherscheme')
    '''

    # --------------------------------------------------------------------------
    #
    @rus.takes  ('Url',
                 rus.optional((str, 'Url')))
    @rus.returns(rus.nothing)
    def __init__(self, url_in=None):
        '''
        Create a new Url object from a string or another Url object.
        '''

        if url_in is None:
            url_in = ''

        self._urlobj = urlparse.urlparse(str(url_in), allow_fragments=True)
        self._renew_url()


    # --------------------------------------------------------------------------
    #
    ##
    @rus.takes  ('Url')
    @rus.returns((rus.nothing, str))
    def __str__ (self):
        '''
        String representation.
        '''

        return self._urlobj.geturl()


    # --------------------------------------------------------------------------
    #
    ##
    @rus.takes  ('Url',
                 ('Url', dict))
    @rus.returns('Url')
    def __deepcopy__(self, memo):
        '''
        Deep copy of a Url
        '''

        return Url(self)


    # --------------------------------------------------------------------------
    #
    ##
    @rus.takes  ('Url')
    @rus.returns(bool)
    def __bool__(self):

        return bool(str(self))


    # --------------------------------------------------------------------------
    #
    ##
    @rus.takes  ('Url',
                 rus.optional(str),
                 rus.optional(str),
                 rus.optional(str),
                 rus.optional((str, int)))
    @rus.returns(str)
    def _make_netloc(self, username, password, hostname, port):
        '''
        helper function to generate netloc string.
        '''

        netloc = ''

        if   username and password: netloc += '%s:%s@' % (username, password)
        elif username             : netloc += '%s@'    %  username
        if   hostname             : netloc += '%s'     %  hostname
        if   port                 : netloc += ':%s'    %  port

        return netloc


    # --------------------------------------------------------------------------
    #
    def _renew_netloc(self, username='', password='', hostname='', port=''):

        newloc = self._make_netloc(username or self._urlobj.username,
                                   password or self._urlobj.password,
                                   hostname or self._urlobj.hostname,
                                   port     or self._urlobj.port)

        self._renew_url(netloc=newloc)


    # --------------------------------------------------------------------------
    #
    def _renew_url(self, scheme='', netloc='', path='',
                         params='', query='',  fragment='', force_path=False):

        # always normalize the path.
        path = self.normpath(path)

        if force_path:
            path = path or '/'

        newurl = urlparse.urlunparse((scheme   or self._urlobj.scheme,
                                      netloc   or self._urlobj.netloc,
                                      path     or self._urlobj.path,
                                      params   or self._urlobj.params,
                                      query    or self._urlobj.query,
                                      fragment or self._urlobj.fragment))

        self._urlobj = urlparse.urlparse(newurl, allow_fragments=True)


    # --------------------------------------------------------------------------
    #
    def normpath(self, path):

        if path:

            # Alas, os.path.normpath removes trailing slashes,
            # so we re-add them.
            if len(path) > 1 and path.endswith('/'):
                trailing_slash = True
            else:
                trailing_slash = False

            path = os.path.normpath(path)

            if trailing_slash and not path.endswith('/'):
                path += '/'

            # os.path.normpath does not normalize for multiple leading slashes
            while path.startswith('//'):
                path = path[1:]

        return path


    # --------------------------------------------------------------------------
    #
    @rus.takes  ('Url',
                 (rus.nothing, str))
    @rus.returns(rus.nothing)
    def set_scheme(self, scheme):
        '''
        Set the URL 'scheme' component.

        :param scheme: The new scheme
        :type  scheme: str

        '''

        self._renew_url(scheme=scheme)


    @rus.takes  ('Url')
    @rus.returns((rus.nothing, str))
    def get_scheme(self):
        '''
        Return the URL 'scheme' component.
        '''

        return self._urlobj.scheme


    scheme = property(get_scheme, set_scheme)
    schema = scheme  # alias, as both terms are used...
    ''' The scheme component.  '''


    # --------------------------------------------------------------------------
    #
    @rus.takes  ('Url',
                 (rus.nothing, str))
    @rus.returns(rus.nothing)
    def set_host(self, hostname):
        '''
        Set the 'hostname' component.

        :param hostname: The new hostname
        :type  hostname: str
        '''

        self._renew_netloc(hostname=hostname)


    @rus.takes  ('Url')
    @rus.returns((rus.nothing, str))
    def get_host(self):
        '''
        Return the URL 'hostname' component.
        '''

        return self._urlobj.hostname


    host = property(get_host, set_host)
    ''' The hostname component.  '''


    # --------------------------------------------------------------------------
    #
    @rus.takes  ('Url',
                 (rus.nothing, str, int))
    @rus.returns(rus.nothing)
    def set_port(self, port):
        '''
        Set the URL 'port' component.

        :param port: The new port
        :type  port: int
        '''

        self._renew_netloc(port=port)


    @rus.takes  ('Url')
    @rus.returns((rus.nothing, int))
    def get_port(self):
        '''
        Return the URL 'port' component.
        '''

        if self._urlobj.port is None:
            return None

        return int(self._urlobj.port)


    port = property(get_port, set_port)
    ''' The port component.  '''


    # --------------------------------------------------------------------------
    #
    @rus.takes  ('Url',
                 (rus.nothing, str))
    @rus.returns(rus.nothing)
    def set_username(self, username):
        '''
        Set the URL 'username' component.

        :param username: The new username
        :type  username: str
        '''

        self._renew_netloc(username=username)


    @rus.takes  ('Url')
    @rus.returns((rus.nothing, str))
    def get_username(self):
        '''
        Return the URL 'username' component.
        '''

        return self._urlobj.username


    username = property(get_username, set_username)
    ''' The username component.  '''


    # --------------------------------------------------------------------------
    #
    @rus.takes  ('Url',
                 (rus.nothing, str))
    @rus.returns(rus.nothing)
    def set_password(self, password):
        '''
        Set the URL 'password' component.

        :param password: The new password
        :type password:  str
        '''

        self._renew_netloc(password=password)


    @rus.takes  ('Url')
    @rus.returns((rus.nothing, str))
    def get_password(self):
        '''
        Return the URL 'username' component.
        '''

        return self._urlobj.password


    password = property(get_password, set_password)
    ''' The password component.  '''


    # --------------------------------------------------------------------------
    #
    @rus.takes  ('Url',
                 (rus.nothing, str))
    @rus.returns(rus.nothing)
    def set_fragment(self, fragment):
        '''
        Set the URL 'fragment' component.

        :param fragment: The new fragment
        :type fragment:  str
        '''

        self._renew_url(fragment=fragment)


    @rus.takes  ('Url')
    @rus.returns((rus.nothing, str))
    def get_fragment(self):
        '''
        Return the URL 'fragment' component.
        '''

        return self._urlobj.fragment


    fragment = property(get_fragment, set_fragment)
    ''' The fragment component.  '''


    # --------------------------------------------------------------------------
    #
    @rus.takes  ('Url',
                 (rus.nothing, str))
    @rus.returns(rus.nothing)
    def set_path(self, path):
        '''
        Set the URL 'path' component.

        :param path: The new path
        :type path:  str
        '''

        self._renew_url(path=path, force_path=True)


    @rus.takes  ('Url')
    @rus.returns((rus.nothing, str))
    def get_path(self):
        '''
        Return the URL 'path' component.
        '''

        path = self._urlobj.path
        path = path.split('?')[0]  # remove query

        return self.normpath(path)


    path = property(get_path, set_path)
    ''' The path component.  '''


    # --------------------------------------------------------------------------
    #
    @rus.takes  ('Url',
                 (rus.nothing, str))
    @rus.returns(rus.nothing)
    def set_query(self, query):
        '''
        Set the URL 'query' component.

        :param query: The new query
        :type query:  str
        '''

        self._renew_url(query=query)


    @rus.takes  ('Url')
    @rus.returns((rus.nothing, str))
    def get_query(self):
        '''
        Return the URL 'query' component.
        '''

        if self._urlobj.query:
            return self._urlobj.query

        if '?' in self._urlobj.path:
            return self._urlobj.path.split('?', 1)[1]  # remove path


    query = property(get_query, set_query)
    ''' The query component.  '''


# --------------------------------------------------------------------





__author__    = "Radical.Utils Development Team (Andre Merzky)"
__copyright__ = "Copyright 2013, RADICAL@Rutgers"
__license__   = "MIT"


import sys
import linecache

_trace_external  = True
_trace_namespace = 'radical'
_trace_logger    = None


# ------------------------------------------------------------------------------
#
'''
  kudos:
  http://www.dalkescientific.com/writings/diary/archive/2005/04/20/\
                                                        tracing_python_code.html


  This module will trace all python function calls, printing each line as it is
  being executed.  It will not print traces for system libraries (i.e. modules
  which are not in the configured namespace), but will indicate when the code
  descents to the system level.

  Python system traces are not following Python's `exec()` call (and
  derivatives), so the resulting trace may be incomplete.   The tracer
  may also fail to step through loaded plugins or adaptors.

  Use like this::

      def my_call(url):

          ru.tracer.trace('radical')

          u = radical.Url(url_str)
          print str(u)

          ru.tracer.untrace()

'''


# ------------------------------------------------------------------------------
#
def _tracer(frame, event, _):

    global _trace_external                               # pylint: disable=W0603

  # if  event == "call":
    if  event == "line":

        filename = frame.f_globals["__file__"]
        lineno   = frame.f_lineno

        if filename.endswith(".pyc") or \
           filename.endswith(".pyo"):
            filename = filename[:-1]

        line = linecache.getline(filename, lineno)

        if _trace_namespace in filename:

            idx  = filename.rindex(_trace_namespace)
            name = filename[idx:]
            if _trace_logger:
                _trace_logger.debug('[trace]: %s %4d %s', name, lineno, line.rstrip())
            else:
                print("%-60s:%4d: %s" % (name, lineno, line.rstrip()))
            _trace_external = False

        else:

            if not _trace_external:
                name = '/'.join(filename.split('/')[-3:])
                if _trace_logger:
                    _trace_logger.debug('[trace]: %s %4d %s', name, lineno, line.rstrip())
                else:
                    print("--> %-56s:%4d: %s" % (name, lineno, line.rstrip()))
            _trace_external = True

    return _tracer


# ------------------------------------------------------------------------------
def trace(namespace='radical', log=None):

    global _trace_namespace                              # pylint: disable=W0603
    global _trace_logger                                 # pylint: disable=W0603

    _trace_namespace = namespace
    _trace_logger    = log

    sys.stdout.write('---------------- start trace [%s]\n' % namespace)
    sys.stdout.flush()
    sys.settrace(_tracer)


# ------------------------------------------------------------------------------
def untrace():

    sys.settrace(None)
    sys.stdout.write('---------------- stop  trace\n')
    sys.stdout.flush()


# ------------------------------------------------------------------------------

